Attribute VB_Name = "Obstacles"
Option Explicit

Public Type Obstacle
  exist As Boolean
  pos As vector
  width As Single
  height As Single
  color As Long
  vel As vector
End Type

Public Const MAXOBSTACLES = 1000
Public numObstacles As Integer
Public Obstacles(MAXOBSTACLES) As Obstacle
Public allowHorizontalObstacleDrift As Boolean
Public allowVerticalObstacleDrift As Boolean
Public makeAllObstaclesBlack As Boolean
Public makeAllObstaclesTransparent As Boolean
Public defaultWidth As Single
Public defaultHeight As Single
Public obstacleDriftRate As Integer
Public obstaclefocus As Integer
Public mousepos As vector
Public mazeCorridorWidth As Integer
Public mazeWallThickness As Integer
Public shapesAreSeeThrough As Boolean
Public leftCompactor As Integer
Public rightCompactor As Integer
Public obstaclesAbsorbShots As Boolean

Public Function DrawHorizontalMaze()
  Dim i As Integer
  Dim j As Integer
  Dim numOfLines As Integer
  Dim Opening As Long
      
  numOfLines = CInt(SimOpts.FieldWidth / (mazeCorridorWidth + mazeWallThickness)) - 1
  For i = 1 To numOfLines
    Opening = Random(0, SimOpts.FieldHeight - mazeCorridorWidth)
    j = NewObstacle(CSng(CSng(i) * CSng(mazeCorridorWidth + mazeWallThickness)), -100, CSng(mazeWallThickness), CSng(Opening))
    If (Opening + mazeCorridorWidth) < SimOpts.FieldHeight + 100 Then
      NewObstacle CSng(CSng(i) * CSng(mazeCorridorWidth + mazeWallThickness)), Opening + CSng(mazeCorridorWidth), CSng(mazeWallThickness), SimOpts.FieldHeight + 100 - Opening - CSng(mazeCorridorWidth)
    End If
  Next i
End Function

Public Function DrawVerticalMaze()
  Dim i As Integer
  Dim j As Integer
  Dim numOfLines As Integer
  Dim Opening As Long
      
  numOfLines = CInt(SimOpts.FieldHeight / (mazeCorridorWidth + mazeWallThickness)) - 1
  For i = 1 To numOfLines
    Opening = Random(0, SimOpts.FieldWidth - mazeCorridorWidth)
    j = NewObstacle(-100, CSng(CSng(i) * CSng(mazeCorridorWidth + mazeWallThickness)), CSng(Opening), CSng(mazeWallThickness))
    If (Opening + mazeCorridorWidth) < SimOpts.FieldWidth + 100 Then
      j = NewObstacle(CSng(Opening + CSng(mazeCorridorWidth)), CSng(i) * CSng(mazeCorridorWidth + mazeWallThickness), SimOpts.FieldWidth + 100 - Opening - CSng(mazeCorridorWidth), CSng(mazeWallThickness))
    End If
  Next i
End Function


Public Function DrawCheckerboardMaze()
  Dim i As Integer
  Dim j As Integer
  Dim k As Integer
  Dim x As Single
  Dim y As Single
  Dim numBlocksAcross As Single
  Dim numBlocksDown As Single
  Dim acrossGap As Single
  Dim downGap As Single
  Dim blockWidth As Single
    
  blockWidth = Min(5000, SimOpts.FieldWidth / 10)
        
  numBlocksAcross = Int(SimOpts.FieldWidth / (blockWidth + mazeCorridorWidth))
  acrossGap = (numBlocksAcross * (blockWidth + mazeCorridorWidth) + mazeCorridorWidth - SimOpts.FieldWidth) / 2
  numBlocksDown = Int(SimOpts.FieldHeight / (blockWidth + mazeCorridorWidth))
  downGap = (numBlocksDown * (blockWidth + mazeCorridorWidth) + mazeCorridorWidth - SimOpts.FieldHeight) / 2
       
  For i = 0 To numBlocksAcross - 1
    For j = 0 To numBlocksDown - 1
      x = CSng(i * blockWidth) + CSng(i + 1#) * CSng(mazeCorridorWidth) - acrossGap
      y = CSng(j * blockWidth) + CSng(j + 1#) * CSng(mazeCorridorWidth) - downGap
      k = NewObstacle(x, y, blockWidth, blockWidth)
    Next j
  Next i
 'allowHorizontalObstacleDrift = True
 ' allowVerticalObstacleDrift = True
 ' obstacleDriftRate = 20
  
End Function

Public Function DrawPolarIceMaze()
  Dim i As Integer
  Dim k As Integer
  Dim blockWidth As Single
  Dim blockHeight As Single
    
  blockWidth = SimOpts.FieldWidth / 2
  blockHeight = SimOpts.FieldHeight / 2
          
  For i = 0 To 8
     k = NewObstacle(blockWidth / 2, blockHeight / 2, blockWidth, blockHeight)
  Next i
  
  allowHorizontalObstacleDrift = True
  allowVerticalObstacleDrift = True
  obstacleDriftRate = 20
  
End Function

Public Function InitTrashCompactorMaze()
  Dim i As Integer
  Dim k As Integer
  Dim blockWidth As Single
  Dim blockHeight As Single
    
  blockWidth = 1000
  blockHeight = SimOpts.FieldHeight * 1.2
          
  leftCompactor = NewObstacle(-blockWidth + 1, SimOpts.FieldHeight * -0.1, blockWidth, blockHeight)
  rightCompactor = NewObstacle(SimOpts.FieldWidth - 1, SimOpts.FieldHeight * -0.1, blockWidth, blockHeight)
  'obstacleDriftRate = 100
  Obstacles(leftCompactor).vel.x = obstacleDriftRate * 0.1
  Obstacles(rightCompactor).vel.x = -obstacleDriftRate * 0.1
  
End Function

Public Function TrashCompactorMove()
  If Obstacles(leftCompactor).pos.x > Obstacles(rightCompactor).pos.x + 400 Then
    Obstacles(leftCompactor).vel.x = -Obstacles(leftCompactor).vel.x
    Obstacles(rightCompactor).vel.x = -Obstacles(rightCompactor).vel.x
  End If
  If Obstacles(leftCompactor).pos.x <= -Obstacles(leftCompactor).width Then
    Obstacles(leftCompactor).vel.x = -Obstacles(leftCompactor).vel.x
    Obstacles(rightCompactor).vel.x = -Obstacles(rightCompactor).vel.x
  End If
End Function


Public Function DrawSpiral()
Dim numOfHorzLines As Integer
Dim numOfVertLines As Integer
Dim numOfLines As Integer
Dim i As Integer
Dim j As Integer

  numOfHorzLines = CInt(SimOpts.FieldHeight / (mazeCorridorWidth + mazeWallThickness)) - 1
  numOfVertLines = CInt(SimOpts.FieldWidth / (mazeCorridorWidth + mazeWallThickness)) - 1
  numOfLines = Min(numOfHorzLines, numOfVertLines)
  If (numOfLines Mod 2) <> 0 Then
    numOfLines = numOfLines - 1
  End If
  For i = 1 To (numOfLines / 2)
    j = NewObstacle(CSng(CSng(i - 1) * CSng(mazeCorridorWidth)), CSng(CSng(i) * CSng(mazeCorridorWidth)), _
                    CSng(SimOpts.FieldWidth - (CSng(mazeCorridorWidth) * (2 * (i - 1) + 1))), CSng(mazeWallThickness))
    j = NewObstacle(CSng(CSng(i) * CSng(mazeCorridorWidth)), CSng(SimOpts.FieldHeight - CSng(CSng(mazeCorridorWidth) * CSng(i))), _
                    CSng(SimOpts.FieldWidth - CSng(mazeCorridorWidth * 2# * CSng(i) - CSng(mazeWallThickness))), CSng(mazeWallThickness))
    j = NewObstacle(CSng(SimOpts.FieldWidth - (CSng(mazeCorridorWidth) * CSng(i))), CSng(CSng(i) * CSng(mazeCorridorWidth)), _
                    CSng(mazeWallThickness), CSng(SimOpts.FieldHeight - (CSng(CSng(mazeCorridorWidth) * CSng(2 * i)))))
    j = NewObstacle(CSng(CSng(i) * CSng(mazeCorridorWidth)), CSng(CSng(i + 1#) * CSng(mazeCorridorWidth)), _
                    CSng(mazeWallThickness), CSng(SimOpts.FieldHeight - CSng(mazeCorridorWidth * CSng(2 * i + 1))))
  Next i
End Function

Public Function NewObstacle(x As Single, y As Single, width As Single, height As Single) As Integer
Dim i As Integer

'  i = 1
'  While Obstacles(i).exist And i < MAXOBSTACLES
'    i = i + 1
'  Wend
  
  If numObstacles + 1 > MAXOBSTACLES Then
    NewObstacle = -1
  Else
    numObstacles = numObstacles + 1
    NewObstacle = numObstacles
    Obstacles(numObstacles).exist = True
    Obstacles(numObstacles).pos.x = x
    Obstacles(numObstacles).pos.y = y
    Obstacles(numObstacles).width = width
    Obstacles(numObstacles).height = height
    If makeAllObstaclesBlack Then
      Obstacles(numObstacles).color = vbBlack
    Else
      Obstacles(numObstacles).color = Rnd(255) * 65536 + Rnd(255) * 255 + Rnd(255) ' Random Color
    End If
   End If
 
 End Function


Public Function DrawObstacles()
  Dim i As Integer
   
Form1.FillStyle = 1

  For i = 1 To numObstacles
    If Obstacles(i).exist Then
      If makeAllObstaclesTransparent Then
        Form1.Line (Obstacles(i).pos.x, Obstacles(i).pos.y)-(Obstacles(i).pos.x + Obstacles(i).width, Obstacles(i).pos.y + Obstacles(i).height), Obstacles(i).color, B
      Else
        Form1.Line (Obstacles(i).pos.x, Obstacles(i).pos.y)-(Obstacles(i).pos.x + Obstacles(i).width, Obstacles(i).pos.y + Obstacles(i).height), Obstacles(i).color, BF
      End If
      If i = obstaclefocus Then
        Form1.Line (Obstacles(i).pos.x - 2, Obstacles(i).pos.y - 2)-(Obstacles(i).pos.x + Obstacles(i).width + 2, Obstacles(i).pos.y + Obstacles(i).height + 2), vbWhite, B
      End If
    End If
  Next i
  
  Form1.FillStyle = 0
End Function

Public Function AddRandomObstacles(n As Integer) As Integer
Dim i As Integer
Dim RandomX As Single
Dim RandomY As Single
Dim RandomWidth As Single
Dim RandomHeight As Single

  If n < 1 Then
    AddRandomObstacles = -1
    Exit Function
  End If
  
  i = 0
  While i <> -1 And n > 0
    RandomX = Rnd * SimOpts.FieldWidth
    RandomY = Rnd * SimOpts.FieldHeight
      
    RandomWidth = Rnd * SimOpts.FieldWidth * defaultWidth
    RandomHeight = Rnd * SimOpts.FieldHeight * defaultHeight
    
    'Shift everything up and left by half the max dimensions then trim to more evenly distribute obstacles across the field
    RandomX = RandomX - SimOpts.FieldWidth * (defaultWidth / 2)
    RandomY = RandomY - SimOpts.FieldHeight * (defaultHeight / 2)
    
    If RandomX < 0 Then RandomX = 0
    If RandomY < 0 Then RandomY = 0
    
    If RandomX + RandomWidth > SimOpts.FieldWidth Then RandomWidth = SimOpts.FieldWidth - RandomX
    If RandomY + RandomHeight > SimOpts.FieldHeight Then RandomHeight = SimOpts.FieldHeight - RandomY
    i = NewObstacle(RandomX, RandomY, RandomWidth, RandomHeight)
    n = n - 1
  Wend
  
  If i = -1 Or n <> 0 Then
    AddRandomObstacles = -1
  Else
    AddRandomObstacles = 0
  End If
  
End Function
Public Function InitObstacles()
Dim i As Integer

  For i = 1 To MAXOBSTACLES
    Obstacles(i).exist = False
  Next i
  numObstacles = 0
End Function
Public Function DeleteAllObstacles()
Dim i As Integer

  For i = 1 To numObstacles
    Obstacles(i).exist = False
  Next i
  numObstacles = 0
End Function

Public Function DeleteObstacle(i As Integer)
  Dim j As Integer
  
  If i < 1 Or i > numObstacles Or numObstacles = 0 Then Exit Function
  For j = i To numObstacles - 1
    Obstacles(j) = Obstacles(j + 1)
  Next j
  Obstacles(numObstacles).exist = False
  numObstacles = numObstacles - 1
  
End Function

Public Function ChangeAllObstacleColor(color As Long)
Dim i As Integer
  
  For i = 1 To numObstacles
    If color < 0 Then
      Obstacles(i).color = Rnd(255) * 65536 + Rnd(255) * 255 + Rnd(255) ' Random Color
    Else
      Obstacles(i).color = color
    End If
  Next i
End Function

Public Function DeleteTenRandomObstacles()
Dim pos As Integer
Dim i As Integer
  
  If numObstacles > 0 Then
    For i = 1 To 10
      DeleteObstacle (Random(1, numObstacles))
    Next i
  End If
 
End Function
 
Public Function MoveObstacles()
  Dim i As Integer
  
    If allowHorizontalObstacleDrift Or allowVerticalObstacleDrift Then DriftObstacles
    If leftCompactor > 0 Or rightCompactor > 0 Then TrashCompactorMove
    
    For i = 1 To numObstacles
      Obstacles(i).pos = VectorAdd(Obstacles(i).pos, Obstacles(i).vel)
      'Keep obstalces from drifting off into space.
      If Obstacles(i).pos.x < -Obstacles(i).width Then
        Obstacles(i).pos.x = -Obstacles(i).width
        Obstacles(i).vel.x = obstacleDriftRate * 0.01
      End If
      If Obstacles(i).pos.y < -Obstacles(i).height Then
        Obstacles(i).pos.y = -Obstacles(i).height
        Obstacles(i).vel.y = obstacleDriftRate * 0.01
      End If
      If Obstacles(i).pos.x > SimOpts.FieldWidth Then
        Obstacles(i).pos.x = SimOpts.FieldWidth
        Obstacles(i).vel.x = -obstacleDriftRate * 0.01
      End If
      If Obstacles(i).pos.y > SimOpts.FieldHeight Then
        Obstacles(i).pos.y = SimOpts.FieldHeight
        Obstacles(i).vel.y = -obstacleDriftRate * 0.01
      End If
      Next i
End Function

Public Function DriftObstacles()
Dim i As Integer

  For i = 1 To numObstacles
    If Obstacles(i).exist And (i <> leftCompactor And i <> rightCompactor) Then
      If allowHorizontalObstacleDrift Then
        Obstacles(i).vel.x = Obstacles(i).vel.x + Random(-obstacleDriftRate, obstacleDriftRate) * Rnd * 0.01
      End If
      If allowVerticalObstacleDrift Then
        Obstacles(i).vel.y = Obstacles(i).vel.y + Random(-obstacleDriftRate, obstacleDriftRate) * Rnd * 0.01
      End If
    End If
  Next i
End Function

Public Function ObstacleCollision(n As Integer, o As Integer) As Boolean
Dim botrightedge As Single
Dim botleftedge As Single
Dim bottopedge As Single
Dim botbottomedge As Single

  ObstacleCollision = False
  
  botrightedge = rob(n).pos.x + rob(n).radius
  botleftedge = rob(n).pos.x - rob(n).radius
  bottopedge = rob(n).pos.y - rob(n).radius
  botbottomedge = rob(n).pos.y + rob(n).radius

  If (botrightedge > Obstacles(o).pos.x) And _
     (botleftedge < Obstacles(o).pos.x + Obstacles(o).width) And _
     (botbottomedge > Obstacles(o).pos.y) And _
     (bottopedge < Obstacles(o).pos.y + Obstacles(o).height) Then
    ObstacleCollision = True
  End If
End Function

Public Function DoShotObstacleCollisions(n As Integer)
Dim i As Integer
  
  With Shots(n)
  For i = 1 To numObstacles
    If Obstacles(i).exist Then
      If .pos.x >= Obstacles(i).pos.x And _
         .pos.x <= Obstacles(i).pos.x + Obstacles(i).width And _
         .pos.y >= Obstacles(i).pos.y And _
         .pos.y <= Obstacles(i).pos.y + Obstacles(i).height Then
         If .opos.x < Obstacles(i).pos.x Or .opos.x > (Obstacles(i).pos.x + Obstacles(i).width) Then
           If obstaclesAbsorbShots Then
             .exist = False
           Else
             .velocity.x = -.velocity.x
           End If
         End If
         If .opos.y < Obstacles(i).pos.y Or .opos.y > (Obstacles(i).pos.y + Obstacles(i).height) Then
            If obstaclesAbsorbShots Then
             .exist = False
           Else
             .velocity.y = -.velocity.y
           End If
         End If
      End If
    End If
  Next i
  End With
End Function

Public Function DoObstacleCollisions(n As Integer)
Dim i As Integer
Dim distleft As Single
Dim distright As Single
Dim distup As Single
Dim distdown As Single
Dim numofcollisions As Integer
Dim LastPush As Integer

numofcollisions = 0
LastPush = 0

Const k As Single = 0.5
Const b As Single = 0.5
With rob(n)
  For i = 1 To numObstacles
    If Obstacles(i).exist Then
      If ObstacleCollision(n, i) Then
        numofcollisions = numofcollisions + 1
        If numofcollisions >= 3 Then
         ' Prevents getting trapped
          .pos.x = .pos.x + 200 * Sgn((SimOpts.TotRunCycle Mod 40) - 20)
          .pos.y = .pos.y + 200 * Sgn((SimOpts.TotRunCycle Mod 50) - 25)
          Exit Function
        End If
        'Push the bot out the closest edge
        distup = (rob(n).pos.y + rob(n).radius) - Obstacles(i).pos.y '- (rob(n).vel.y / 2)
        distdown = Obstacles(i).pos.y + Obstacles(i).height - (rob(n).pos.y - rob(n).radius) '- (rob(n).vel.y / 2)
        distleft = (rob(n).pos.x + rob(n).radius) - Obstacles(i).pos.x '- (rob(n).vel.x / 2)
        distright = Obstacles(i).pos.x + Obstacles(i).width - (rob(n).pos.x - rob(n).radius) '- (rob(n).vel.x / 2)
               
        If (Min(distleft, distright) < Min(distup, distdown) And _
           (LastPush <> 1 And LastPush <> 2)) Or _
           (LastPush = 3 Or LastPush = 4) Then
          'Push out left or right
          If ((distleft <= distright) Or _
             (Obstacles(i).pos.x + Obstacles(i).width) >= SimOpts.FieldWidth) And _
             (Obstacles(i).pos.x > 0) Then
          
            If rob(n).pos.x - rob(n).radius < Obstacles(i).pos.x Then
              .pos.x = Obstacles(i).pos.x - rob(n).radius
              .ImpulseRes.x = .ImpulseRes.x + .vel.x * b
               touch n, .pos.x + .radius, .pos.y ' Update hit senses, right side
            Else
              .ImpulseRes.x = .ImpulseRes.x + distleft * k
              If .Fixed Then .pos = VectorSub(.pos, .ImpulseRes) ' force .fixed guys to move without changing their fixedness
            End If
            LastPush = 1
          Else
            If rob(n).pos.x + rob(n).radius > Obstacles(i).pos.x + Obstacles(i).width Then
              .pos.x = Obstacles(i).pos.x + Obstacles(i).width + rob(n).radius
              .ImpulseRes.x = .ImpulseRes.x + .vel.x * b
              touch n, .pos.x - .radius, .pos.y ' Update hit senses, left side
            Else
              .ImpulseRes.x = .ImpulseRes.x - distright * k
               If .Fixed Then .pos = VectorSub(.pos, .ImpulseRes) ' force .fixed guys to move without changing their fixedness
            End If
            LastPush = 2
          End If
        Else
          'Push out up or down
          If ((distup <= distdown) Or _
             (Obstacles(i).pos.y + Obstacles(i).height) >= SimOpts.FieldHeight) And _
             (Obstacles(i).pos.y > 0) Then
            If rob(n).pos.y - rob(n).radius < Obstacles(i).pos.y Then
              .pos.y = Obstacles(i).pos.y - rob(n).radius
              .ImpulseRes.y = .ImpulseRes.y + .vel.y * b
              touch n, .pos.x, .pos.y + .radius  ' Update hit senses, bottom
            Else
              .ImpulseRes.y = .ImpulseRes.y + distup * k
              If .Fixed Then .pos = VectorSub(.pos, .ImpulseRes) ' force .fixed guys to move without changing their fixedness
            End If
            LastPush = 3
          Else
            If rob(n).pos.y + rob(n).radius > Obstacles(i).pos.y + Obstacles(i).height Then
              .pos.y = Obstacles(i).pos.y + Obstacles(i).height + rob(n).radius
              .ImpulseRes.y = .ImpulseRes.y + .vel.y * b
              touch n, .pos.x, .pos.y - .radius  ' Update hit senses, bottom
            Else
              .ImpulseRes.y = .ImpulseRes.y - distdown * k
              If .Fixed Then .pos = VectorSub(.pos, .ImpulseRes) ' force .fixed guys to move without changing their fixedness
            End If
            
           LastPush = 4
          End If
        End If
       
       ' If VectorMagnitude(.ImpulseRes) > VectorMagnitude(.vel) Then
       '   .ImpulseRes = VectorScalar(.ImpulseRes, (VectorMagnitude(.vel) / VectorMagnitude(.ImpulseRes)) * 0.99)
       ' End If
      End If
    End If
  Next i
  
 ' If numofcollisions > 2 Then
    'Give up and just get them out of there
 '   .pos.x = Rnd * SimOpts.FieldWidth
 '   .pos.y = Rnd * SimOpts.FieldHeight
 ' End If
    'ImpulseRes.y = .ImpulseRes.y - SimOpts.MaxVelocity * (Rnd(1) * -2 + 1) * Rnd(1)
'    .ImpulseRes.x = .ImpulseRes.x - SimOpts.MaxVelocity * (Rnd(1) * -2 + 1) * Rnd(1)
'  End If

  End With
End Function

Public Function whichobstacle(x As Single, y As Single) As Integer
  Dim t As Integer
  whichobstacle = 0
  For t = numObstacles To 1 Step -1
    If Obstacles(t).exist Then
      If x >= Obstacles(t).pos.x And x <= Obstacles(t).pos.x + Obstacles(t).width And _
         y >= Obstacles(t).pos.y And y <= Obstacles(t).pos.y + Obstacles(t).height Then
         whichobstacle = t
         Exit Function
      End If
    End If
  Next t
End Function
